import abc
# from flask_scrapinghub.chainspider.spider_chain import SpiderChain


class SpiderBase(abc.ABC):

    def __init__(self, project_id, spider_id):
        self.project_id = project_id
        self.spider_id = spider_id

    @abc.abstractmethod
    def see_black(self, current_index, chain):
        """

        :param chain:
        :param current_index: int. The self position in chain
        """
        return NotImplemented

    @abc.abstractmethod
    def see_red(self, current_index, chain):
        """

        :param chain:
        :param current_index: int. The self position in chain
        """
        return NotImplemented

    @abc.abstractmethod
    def see_green(self, current_index, chain):
        """

        :param chain:
        :param current_index: int. The self position in chain
        """
        return NotImplemented

    @abc.abstractmethod
    def see_yellow(self, current_index, chain):
        """

        :param chain:
        :param current_index: int. The self position in chain
        """
        return NotImplemented



import abc
from flask_scrapinghub.chainspider.spider_base import SpiderBase
from flask_scrapinghub.spider_console import SpiderConsole
from flask_scrapinghub.utils.color import Color
from datetime import datetime
import time


class SpiderChain:

    def __init__(self, head: SpiderBase, console: SpiderConsole):
        self.__console = console
        if not self.__console.check_spider_color(head.project_id, head.spider_id):
            # todo: 所輸入乃 console 記錄中不存在的 spider
            raise Exception
        self.__head = head
        self.__start_execution_time = None
        self.__current_index = 0
        self.__chain = []
        self.add(self.__head)

    def reset(self, head: SpiderBase):
        """
        note:
            初始化 chain
        """
        if not self.__console.check_spider_color(head.project_id, head.spider_id):
            # todo: 所輸入乃 console 記錄中不存在的 spider
            raise Exception
        self.__head = head
        self.__chain = []

    def add(self, mate: SpiderBase):
        """
        note:
            在 SpiderChain中添加 繼承至SpiderBase的物件。
            若當前為空 Chain，則加入的對象會成為 head，且會要求後來加入者與 head 為同類型;
            若當前為空 Chain，則加入對象為 mate，並且會檢查是否與head同類型、且檢查是否有相同 mate 的存在。。

        :type mate: SpiderBase
        """
        # if not isinstance(mate, type(self.head)):
        #     # todo: 加入與 head 不同類型的 mate
        #     raise Exception
        for m in self.__chain:
            if mate.project_id == m.project_id and mate.spider_id == m.spider_id:
                # todo: 已存在相同 mate
                raise Exception
        if not self.__console.check_spider_color(mate.project_id, mate.spider_id):
            # todo: 所輸入乃 console 記錄中不存在的 spider
            raise Exception
        self.__chain.append(mate)

    @property
    def length(self):
        return len(self.__chain)

    def start(self):
        if not self.shall_start(self.__head):
            # todo: 未通過 shall start 檢測
            return
        self.__start_execution_time = datetime.now()
        self.__current_index = 0
        self.execute_node()

    def execute_node(self):
        node = self.__chain[self.__current_index]
        color = self.__console.check_spider_color(node.project_id, node.spider_id)
        if color is Color.Red:
            node.see_red(self.__current_index, self)
        elif color is Color.YELLOW:
            node.see_yellow(self.__current_index, self)
        elif color is Color.BLACK:
            node.see_black(self.__current_index, self)
        else:
            node.see_green(self.__current_index, self)

    def execute_next_node(self):
        self.__current_index = 0 if self.__current_index + 1 >= len(self.__chain) \
                                 else self.__current_index + 1
        self.execute_node()

    def try_execute_again_after_interval(self, check_interval_sec: int = 30):
        time.sleep(check_interval_sec)
        self.execute_node()

    def take_items(self, mate: SpiderBase):
        items = self.__console.take_items_from_spider(mate.project_id, mate.spider_id)
        return items

    def run_node(self, mate: SpiderBase):
        execution = self.__console.run_spider(mate.project_id, mate.spider_id)
        return execution

    def shut_all(self):
        for node in self.__chain:
            execution = self.__console.stop_spider(node.project_id, node.spider_id)
            if not execution:
                # todo: project_id 或 spider_id 異常
                raise Exception

    def resume_all(self):
        for node in self.__chain:
            execution = self.__console.resume_spider(node.project_id, node.spider_id)
            if not execution:
                # todo: project_id 或 spider_id 異常
                raise Exception

    def check_all(self):
        status_list = list()
        for node in self.__chain:
            color = self.__console.check_spider_color(node.project_id, node.spider_id)
            if not color:
                # todo: project_id 或 spider_id 異常
                raise Exception
            status_list.append(color)
        return status_list

    def execution_total_time(self):
        """

        :return: todo: 回傳值
        """
        if not self.__start_execution_time:
            return None
        return datetime.now() - self.__start_execution_time

    @abc.abstractmethod
    def shall_start(self, head):
        return True



from flask_scrapinghub.chainspider.spider_base import SpiderBase
from flask_scrapinghub.chainspider.spider_chain import SpiderChain
from scrapinghub.client.items import Items
from flask_scrapinghub.utils.color import Color
import abc


class SpiderNode(SpiderBase):

    def __init__(self, project_id, spider_id, check_interval_sec=30):
        self.check_interval_sec = check_interval_sec
        super().__init__(project_id, spider_id)

    def see_black(self, current_index, chain: SpiderChain):
        not_meet_count = self.__not_meet_color_count(chain, Color.BLACK)
        if not_meet_count == 0:
            chain.start()
        else:
            chain.shut_all()
            chain.try_execute_again_after_interval(self.check_interval_sec)

    def see_red(self, current_index, chain: SpiderChain):
        not_meet_count = self.__not_meet_color_count(chain, Color.Red)
        if not_meet_count == 0:
            items = chain.take_items(self)
            self.feed_items(items)
            if current_index + 1 == chain.length:
                chain.resume_all()
            else:
                chain.execute_next_node()
        else:
            chain.try_execute_again_after_interval(self.check_interval_sec)

    def see_green(self, current_index, chain: SpiderChain):
        not_meet_count = self.__not_meet_color_count(chain, Color.GREEN)
        if current_index == 0:
            if not_meet_count == 0:
                execution = chain.run_node(self)
                chain.execute_next_node()
            else:
                chain.try_execute_again_after_interval(self.check_interval_sec)
        else:
            execution = chain.run_node(self)
            chain.execute_next_node()

    def see_yellow(self, current_index, chain: SpiderChain):
        not_meet_count = self.__not_meet_color_count(chain, Color.Red)
        chain.try_execute_again_after_interval(self.check_interval_sec)
        # if not_meet_count == 0:
        #     chain.execute_next_node()
        # else:
        #     chain.try_execute_again_after_interval(self.check_interval_sec)

    @classmethod
    def __not_meet_color_count(cls, chain: SpiderChain, specific_color: Color):
        all_color = chain.check_all()
        not_meet_count = 0
        for color in all_color:
            if color is not specific_color:
                not_meet_count += 1
        return not_meet_count

    @abc.abstractmethod
    def feed_items(self, items):
        return NotImplemented



from scrapinghub import ScrapinghubClient
from flask_scrapinghub.utils.tag import Tag
from flask_scrapinghub.utils.color import Color
from flask_scrapinghub.utils.execution import Execution


class SpiderSettings:

    def __init__(self, apikey: str, projects: list):
        self.api_key = apikey
        self.projects = projects


class SpiderSig:

    def __init__(self, spider_id):
        self.id = spider_id
        self.tag = Tag.STAND_BY
        self.__job_key = None

    def set_job_key(self, job_key):
        self.__job_key = job_key

    def release_job_key(self):
        self.__job_key = None

    def job_key(self):
        return self.__job_key

    def to_standby(self):
        self.tag = Tag.STAND_BY

    def to_running(self):
        self.tag = Tag.RUNNING

    def to_finished(self):
        self.tag = Tag.FINISHED

    def to_stop(self):
        self.tag = Tag.STOP


class ScrapinghubService:

    def __init__(self, api_key):
        self.client = ScrapinghubClient(api_key)

    def __get_project(self, project_id):
        """

        :param project_id: string project id
        :return: Project 實體 (class: ~scrapinghub.client.projects.Project)
        """
        return self.client.projects.get(project_id)

    def __get_jobs(self, project_id):
        """

        :param project_id: string project id
        :return: Jobs 實體 (class: ~scrapinghub.client.jobs.Jobs)
        """
        return self.client.projects.get(project_id).jobs

    def run_spider_job(self, project_id, unit: SpiderSig):
        """

        :param project_id: string project id
        :param unit: unit of project (class: SpiderUnit).
        :return: a string of executed job key
        """
        job = self.__get_jobs(project_id).run(
                                      spider=unit.id,
                                      units=None, priority=None, meta=None,
                                      add_tag=None, job_args=None, job_settings=None, cmd_args=None,
                                      environment=None)
        return job.key

    def cancel_job(self, project_id, job_key: str):
        """

        :param project_id: string project id
        :param job_key: a string of executed job key
        """
        self.__get_jobs(project_id).get(job_key).cancel()

    def retrieve_unit_state(self, project_id, unit: SpiderSig):
        """

        :param project_id: string project id
        :param unit: unit of project (class: SpiderSig).
        :return: a string of unit job state
        """
        state = self.__get_jobs(project_id).get(unit.job_key()).metadata.get('state') \
            if unit.job_key else None
        return state

    def retrieve_unit_items(self, project_id, unit: SpiderSig):
        """

        :param project_id: string project id
        :param unit: unit of project (class: SpiderSig).
        :return: items (class:`~scrapinghub.client.items.Items` resource object).
        """

        items = self.__get_jobs(project_id).get(unit.job_key()).items \
            if unit.job_key else None

        return items

    def retrieve_unit_logs(self, project_id, unit: SpiderSig):
        """

        :param project_id: string project id
        :param unit: unit of project (class: SpiderSig).
        :return: logs (class:`~scrapinghub.client.logs.Logs` resource object).
        """
        logs = self.__get_project(project_id).jobs.get(unit.job_key).logs \
            if unit.job_key else None
        return logs

    def get_project_spiders_list(self, project_id):
        """

        :param project_id: string project id
        :return: a list of spiders for a project
        [{'id': 'spider1', 'tags': [], 'type': 'manual', 'version': '123'},
         {'id': 'spider2', 'tags': [], 'type': 'manual', 'version': '123'}]
        """
        spiders_list = self.__get_project(project_id).spiders.list()
        return spiders_list


class SpiderSignalsCenter:

    def __init__(self):
        self.__spider_signals = {}

    @classmethod
    def build_signals_center(cls, projects_list, center_client: ScrapinghubService):
        sig_center = SpiderSignalsCenter()
        for project_id in projects_list:
            project_spiders_list = center_client.get_project_spiders_list(project_id)
            sig_center.__spider_signals[project_id] = {}
            sig_center.__spider_signals[project_id]["spiders"] = list()
            for spider_item in project_spiders_list:
                sig_center.__spider_signals[project_id]["spiders"].append(SpiderSig(spider_item['id']))
        return sig_center

    def __get_spider(self, project_id, spider_id):
        """ 取得 spider

        :param project_id: string
        :param spider_id: string
        :return: self.spider_signals 中 project_id底下的spiders陣列中的spider

        ex: {'name': 'spider1',
             'status': (Tag type)}
        """
        project = self.__spider_signals.get(project_id, None)
        if not project:
            return None
        spiders = project.get("spiders", None)
        if not spiders or len(spiders) == 0:
            return None
        for spider in spiders:
            if spider.id == spider_id:
                return spider
        return None

    def try_get_sig(self, project_id, spider_id):
        spider_sig = self.__get_spider(project_id, spider_id)
        if not spider_sig:
            return None
        return spider_sig

    def tag_stand_by(self, project_id, spider_id):
        spider_sig = self.__get_spider(project_id, spider_id)
        if not spider_sig:
            raise Exception
        spider_sig.to_standby()

    def tag_running(self, project_id, spider_id):
        spider_sig = self.__get_spider(project_id, spider_id)
        if not spider_sig:
            raise Exception
        spider_sig.to_running()

    def tag_finished(self, project_id, spider_id):
        spider_sig = self.__get_spider(project_id, spider_id)
        if not spider_sig:
            raise Exception
        spider_sig.to_finished()

    def tag_stop(self, project_id, spider_id):
        spider_sig = self.__get_spider(project_id, spider_id)
        if not spider_sig:
            raise Exception
        spider_sig.to_stop()


class SpiderExecutor:

    def __init__(self, settings: SpiderSettings):
        self.client = ScrapinghubService(settings.api_key)

    def check_spider_color(self, project_id, sig: SpiderSig):
        """

        :param project_id:  a string of project id.
        :param sig: (class: `SpiderSig`).
        :return: color (class:`~flask_scrapinghub.utils.color.Color`) or None.
        """
        tag = sig.tag
        if tag is Tag.FINISHED:
            return Color.Red
        if tag is Tag.STAND_BY:
            return Color.GREEN
        state = self.client.retrieve_unit_state(project_id, sig)
        if tag is Tag.RUNNING:
            if state == "finished":
                sig.to_finished()
                return Color.Red
            else:
                return Color.YELLOW
        if tag is Tag.STOP:
            if state == "running":
                return Color.BLACK
            else:
                sig.to_standby()
                sig.release_job_key()
                return Color.GREEN

    def run_spider(self, project_id, sig: SpiderSig):
        """

        :param project_id:  a string of project id.
        :param sig: (class: `SpiderSig`).
        :return: execution (class:`~flask_scrapinghub.utils.execution.Execution`) or None.
        """
        tag = sig.tag
        if tag is Tag.FINISHED:
            return Execution.FAIL
        if tag is Tag.RUNNING:
            return Execution.FAIL
        if tag is Tag.STOP:
            return Execution.FAIL
        if tag is Tag.STAND_BY:
            job_key = self.client.run_spider_job(project_id, sig)
            sig.to_running()
            sig.set_job_key(job_key)
            return Execution.SUCCESS

    def stop_spider(self, project_id, sig: SpiderSig):
        """

        :param project_id:  a string of project id.
        :param sig: (class: `SpiderSig`).
        :return: execution (class:`~flask_scrapinghub.utils.execution.Execution`) or None.
        """
        tag = sig.tag
        if tag is Tag.FINISHED:
            return Execution.FAIL
        if tag is Tag.STAND_BY:
            return Execution.FAIL
        if tag is Tag.STOP:
            return Execution.FAIL
        if tag is Tag.RUNNING:
            self.client.cancel_job(project_id, sig.job_key())
            sig.to_stop()
            return Execution.SUCCESS

    def take_items_from_spider(self, project_id, sig: SpiderSig):
        """

        :param project_id:  a string of project id.
        :param sig: (class: `SpiderSig`).
        :return: items (class:`~scrapinghub.client.items.Items` resource object).
        """

        tag = sig.tag
        if tag is Tag.STAND_BY:
            return None
        if tag is Tag.STOP:
            return None
        if tag is Tag.RUNNING:
            return None
        if tag is Tag.FINISHED:
            return self.client.retrieve_unit_items(project_id, sig)

    def resume_spider(self, sig: SpiderSig):
        """

        :param sig: (class: `SpiderSig`).
        :return: execution (class:`~flask_scrapinghub.utils.execution.Execution`) or None.
        """
        tag = sig.tag
        if tag is Tag.STAND_BY:
            return Execution.FAIL
        if tag is Tag.STOP:
            return Execution.FAIL
        if tag is Tag.RUNNING:
            return Execution.FAIL
        if tag is Tag.FINISHED:
            sig.to_standby()
            return Execution.SUCCESS


class SpiderConsole:

    def __init__(self, settings: SpiderSettings):
        self.__executor = SpiderExecutor(settings)
        self.__sig_center = SpiderSignalsCenter\
            .build_signals_center(settings.projects, self.__executor.client)

    def check_spider_color(self, project_id, spider_id):
        sig = self.__sig_center.try_get_sig(project_id, spider_id)
        if not sig:
            return None
        return self.__executor.check_spider_color(project_id, sig)

    def run_spider(self, project_id, spider_id):
        sig = self.__sig_center.try_get_sig(project_id, spider_id)
        if not sig:
            return None
        return self.__executor.run_spider(project_id, sig)

    def stop_spider(self, project_id, spider_id):
        sig = self.__sig_center.try_get_sig(project_id, spider_id)
        if not sig:
            return None
        return self.__executor.stop_spider(project_id, sig)

    def take_items_from_spider(self, project_id, spider_id):

        sig = self.__sig_center.try_get_sig(project_id, spider_id)
        if not sig:
            return None

        return self.__executor.take_items_from_spider(project_id, sig)

    def resume_spider(self, project_id, spider_id):
        sig = self.__sig_center.try_get_sig(project_id, spider_id)
        if not sig:
            return None
        return self.__executor.resume_spider(sig)



from flask_scrapinghub.utils.color import Color
import abc
import time


class SpiderBase(object):

    def __init__(self, project_id, spider_id, tag=None):
        self.project_id = project_id
        self.spider_id = spider_id
        self.tag = tag

    def react(self, color, next_round, client):
        if not Color:
            return
        if type(color) != Color:
            raise Exception('spider only react to color')
        if color == Color.Red:
            self.see_red(next_round, client)
        elif color == Color.YELLOW:
            self.see_yellow(next_round, client)
        elif color == Color.BLACK:
            self.see_black(next_round, client)
        else:
            self.see_green(next_round, client)

    def add_tag(self, tag):
        self.tag = tag

    def shall_start(self, client):
        return True

    @abc.abstractmethod
    def see_black(self, next_round, client):
        return NotImplemented

    @abc.abstractmethod
    def see_red(self, next_round, client):
        return NotImplemented

    @abc.abstractmethod
    def see_green(self, next_round, client):
        return NotImplemented

    @abc.abstractmethod
    def see_yellow(self, next_round, client):
        return NotImplemented


class Butterfly(SpiderBase):

    def shall_start(self, client):
        return True

    def __init__(self, project_id, spider_id):
        super().__init__(project_id, spider_id)

    def see_black(self, next_round, client):
        pass

    def see_red(self, next_round, client):
        pass

    def see_green(self, next_round, client):
        print("I am Butterfly  with green")
        time.sleep(5)
        if next_round != 0:
            client.wake_spider()
        # client.run_spider(self.project_id, self.spider_id)

    def see_yellow(self, next_round, client):
        print("I am Butterfly  with yellow")


class Bee(SpiderBase):

    def __init__(self, project_id, spider_id):
        super().__init__(project_id, spider_id)

    def shall_start(self, client):
        return True

    def see_black(self, next_round, client):
        pass

    def see_red(self, next_round, client):
        pass

    def see_green(self, next_round, client):
        print("I am Bee with green")
        time.sleep(5)
        if next_round != 0:
            client.wake_spider()
        # client.run_spider(self.project_id, self.spider_id)

    def see_yellow(self, next_round, client):
        print("I am Butterfly  with yellow")



from flask_scrapinghub.chainspider.spider_base import SpiderBase
from flask_scrapinghub.utils.color import Color
from flask_scrapinghub.spider_console import SpiderSettings, SpiderConsole
from scrapinghub import ScrapinghubClient
from scrapinghub.client.exceptions import DuplicateJobError, BadRequest
from requests import ConnectionError
from flask_scrapinghub.chainspider.spider_chain import SpiderChain
import logging
import json

logger = logging.getLogger(__name__)


class FlaskScrapinghub(object):

    def __init__(self, config_prefix="SCRAPINGHUB"):
        self.settings: SpiderSettings = None
        self.console: SpiderConsole = None
        self.spider_chain: SpiderChain = None
        self.config_prefix = config_prefix

    def init_app(self, app):
        setup_file = app.config.get(f"{self.config_prefix}_SETUP_FILE", 'setup.json')
        with open(setup_file, 'r') as json_file:
            settings = json.load(json_file)
        apikey = settings.get('api-key', None)
        projects = settings.get("projects", None)
        self.settings = SpiderSettings(apikey, projects)
        self.console = SpiderConsole(self.settings)
        return self

    def init_spider_chain(self, head: SpiderBase, spider_chain_type: type = None):
        if not self.console:
            raise Exception
        if spider_chain_type:
            try:
                spider_chain = spider_chain_type(head=head,
                                             console=self.console)
            except TypeError:
                raise Exception("spider_chain_type parameters 傳入型態有誤")
        else:
            spider_chain = SpiderChain(head=head,
                                       console=self.console)
        if not isinstance(spider_chain, SpiderChain):
            raise Exception
        self.spider_chain = spider_chain
        return self

    # def __getattr__(self, name):
    #     return getattr(self.console, name)

    # def __init__(self, app=None, config_prefix="SCRAPINGHUB", spiders_mode=False, spiders_tag='none', **kwargs):
    #     # 爬蟲模式
    #     self.spiders_mode = spiders_mode
    #     self.spiders_tag = spiders_tag
    #     self.spiders_mode_index = -1
    #     self.spiders = None
    #     # 基本配置
    #     self.client = None
    #     self.setup = None
    #     self.provider_kwargs = kwargs
    #     self.config_prefix = config_prefix
    #     self.spider_signals = {}
    #     self.initialized = False
    #     if app is not None:
    #         self.init_app(app)

    # def init_app(self, app, **kwargs):
    #     if self.spiders_mode:
    #         self.spiders = []
    #         self.spiders_mode_index = 0
    #     logger.info(f"Spider Mode: {self.spiders_mode}")
    #     # 讀取 setup_development.json (預設命名 SCRAPINGHUB_SETUP_FILE)
    #     setup_file = app.config.get(f"{self.config_prefix}_SETUP_FILE", 'setup.json')
    #     # 取得 api-key，以創建client
    #     with open(setup_file, 'r') as json_file:
    #         self.setup = json.load(json_file)
    #     apikey = self.setup.get('api-key', None)
    #     if not apikey:
    #         raise SyntaxError()
    #     self.client = ScrapinghubClient(apikey)
    #
    #     # app 讀入 client
    #     if not hasattr(app, "extensions"):
    #         app.extensions = {}
    #     if app.extensions.get(self.config_prefix.lower(), None):
    #         logger.warning(
    #             f"Another FlaskScrapinghub instance using the same prefix: \"{self.config_prefix.lower()}\"! ")
    #     app.extensions[self.config_prefix.lower()] = self
    #     self.provider_kwargs.update(kwargs)
    #     # 初始化 spider_signals
    #     self.init_spider_signals()
    #     self.initialized = True

    # def make_sure_initialized(self):
    #     if not self.initialized:
    #         raise Exception('Has not initialized Error')

    # def init_spider_signals(self):
    #     projects = self.setup.get("projects", None)
    #     if not projects or len(projects) == 0:
    #         self.spider_signals = {}
    #         return
    #     for project in projects:
    #         project_instance = self.client.get_project(project)
    #         self.spider_signals[project] = {}
    #         self.spider_signals[project]["jobs"] = project_instance.jobs
    #         self.spider_signals[project]["spiders"] = []
    #         for spider in project_instance.spiders.list():
    #             self.spider_signals[project]["spiders"].append({
    #                 "name": spider['id'],
    #                 "status": self.tag_stand_by(),
    #                 "jobs": project_instance.jobs,
    #                 "items": None,
    #                 "logs": None,
    #                 "job": None
    #             })

    # def check_and_get_spider(self, project_id, spider_id):
    #     self.make_sure_initialized()
    #     project = self.spider_signals.get(project_id, None)
    #     # -----若 project_id不存在-----
    #     if not project:
    #         return None
    #     spiders = project.get("spiders", None)
    #     # -----若 project_id不存在-----
    #     if not spiders or len(spiders) == 0:
    #         return None
    #     # -----若 project中 spider不存在-----
    #     if not any(spider.get("name", None) == spider_id for spider in spiders):
    #         return None
    #     specific_spider = None
    #     for spider in self.spider_signals[project_id]['spiders']:
    #         if spider["name"] == spider_id:
    #             specific_spider = spider
    #     return specific_spider
    #
    # def view_spider_info(self, project_id, spider_id):
    #     self.make_sure_initialized()
    #     spider = self.check_and_get_spider(project_id, spider_id)
    #     # -----所指定的 spider不存在-----
    #     if not spider:
    #         return None
    #     name = spider.get('name', None)
    #     status = spider.get('status', None)
    #     jobs = spider.get('jobs', None)
    #     job = spider.get('job', None)
    #     job_status = self.get_job_status(jobs, job) if job and jobs else None
    #     info = {
    #         'name': name,
    #         'spider_status': status,
    #         'job': job,
    #         'job_status': job_status
    #     }
    #     return info
    #
    # def check_spider(self, project_id, spider_id):
    #     self.make_sure_initialized()
    #     spider = self.check_and_get_spider(project_id, spider_id)
    #     # -----所指定的 spider不存在-----
    #     if not spider:
    #         return None
    #     # -----若 spider已完成-----
    #     if spider['status'] == self.tag_finished():
    #         return Color.Red
    #     # -----若 spider待命中-----
    #     if spider['status'] == self.tag_stand_by():
    #         return Color.GREEN
    #     # -----若 spider進行中-----
    #     if spider['status'] == self.tag_running():
    #         jobs = spider.get('jobs', None)
    #         job = spider.get('job', None)
    #         if not job or not jobs:
    #             raise Exception('spider停止，但無執行中工作，程式邏輯有誤')
    #         spider_status = self.get_job_status(jobs, job)
    #         # 若已結束工作
    #         if spider_status == "finished":
    #             spider['status'] = self.tag_finished()
    #             return Color.Red
    #         return Color.YELLOW
    #     # -----若 spider停止中-----
    #     if spider['status'] == self.tag_stop():
    #         jobs = spider.get('jobs', None)
    #         job = spider.get('job', None)
    #         if not job or not jobs:
    #             raise Exception('spider停止，但無最後工作，程式邏輯有誤')
    #         spider_status = self.get_job_status(jobs, job)
    #         # 若最後工作仍在運行中
    #         if spider_status == "running":
    #             return Color.BLACK
    #         # 若最後工作已處於非運行狀態
    #         else:
    #             spider['status'] = self.tag_stand_by()
    #             spider['items'] = None
    #             spider['logs'] = None
    #             spider['job'] = None
    #             return Color.GREEN
    #
    # def run_spider(self, project_id, spider_id, units=1, priority=4, tags=None,
    #                current_try_time=None, max_try_times=None):
    #
    #     self.make_sure_initialized()
    #     spider = self.check_and_get_spider(project_id, spider_id)
    #     # -----所指定的 spider不存在-----
    #     if not spider:
    #         return Execution.NO_SPIDER
    #     # -----若 spider已完成-----
    #     if spider['status'] == self.tag_finished():
    #         logger.warning(f"[{project_id}/{spider_id}] 已完成，需先進行資料轉移始可重新啟動")
    #         return Execution.FAIL
    #     # -----若 spider進行中-----
    #     if spider['status'] == self.tag_running():
    #         logger.warning(f"[{project_id}/{spider_id}] 進行中，無法啟動，請先停止爬蟲")
    #         return Execution.FAIL
    #     # -----若 spider停止中-----
    #     if spider['status'] == self.tag_stop():
    #         logger.warning(f"[{project_id}/{spider_id}] 停止中，再次確認當前狀態")
    #         light = self.check_spider(project_id, spider_id)
    #         if light != Color.GREEN:
    #             return Execution.FAIL
    #         else:
    #             logger.info(f"[{project_id}/{spider_id}] 已恢復，將自動嘗試啟動爬蟲")
    #             return self.run_spider(project_id, spider_id, units, priority, tags)
    #
    #     # -----若 spider待命中-----
    #     if spider['status'] == self.tag_stand_by():
    #         logger.info(f"[{project_id}/{spider_id}] 待命中，即將啟動，開始爬蟲作業")
    #
    #     try:
    #         job = self.client.get_project(project_id) \
    #             .jobs.run(spider=spider_id,
    #                       units=units,
    #                       priority=priority,
    #                       tags=tags)
    #     except DuplicateJobError:
    #         logger.warning(f"[{project_id}/{spider_id}] 相同爬蟲正在工作中，當前操作無效")
    #         spider['status'] = self.tag_running()
    #         running_summary = [i.get('summary') for i in spider['jobs'].summary() if i.get('name') == 'running'][0]
    #         if running_summary:
    #             found_spider_status = [s for s in running_summary if s.get('spider') == spider_id]
    #         else:
    #             found_spider_status = None
    #         if not found_spider_status:
    #             logger.warning(f"[{project_id}/{spider_id}] 當前無無任何爬蟲工作中記錄，邏輯有誤")
    #         elif len(found_spider_status) == 0:
    #             logger.warning(f"[{project_id}/{spider_id}] 查無此爬蟲並未工作中，邏輯有誤")
    #         else:
    #             spider_status = found_spider_status[0]
    #             key = spider_status.get('key')
    #             job = spider['jobs'].get(key)
    #             spider['job'] = key
    #             spider['items'] = job.items
    #             spider['logs'] = job.logs
    #             if key == f"{project_id}/spider['job']":
    #                 logger.info(f"[{project_id}/{spider_id}] 爬蟲存在相關工作紀錄，原任務繼續進行")
    #             else:
    #                 logger.warning(f"[{project_id}/{spider_id}] 爬蟲無相關紀錄，將立即取消該工作")
    #                 self.stop_spider(project_id, spider_id)
    #         return Execution.FAIL
    #     except BadRequest or ConnectionError:
    #         if not max_try_times:
    #             max_try_times = 5
    #         if not current_try_time:
    #             current_try_time = 1
    #         if current_try_time <= max_try_times:
    #             logger.error(f"[{project_id}/{spider_id}] 當前爬蟲請求操作失敗，嘗試次數已達上限，宣告失敗")
    #             return Execution.FAIL
    #         logger.warning(f"[{project_id}/{spider_id}] 當前爬蟲請求操作失敗，嘗試再次發送")
    #         return self.run_spider(project_id, spider_id, units, priority, tags, current_try_time + 1, max_try_times)
    #
    #     if spider["name"] == spider_id:
    #         spider['status'] = self.tag_running()
    #         spider['items'] = job.items
    #         spider['logs'] = job.logs
    #         spider['job'] = job.key
    #     return Execution.SUCCESS
    #
    # def stop_spider(self, project_id, spider_id):
    #
    #     self.make_sure_initialized()
    #     spider = self.check_and_get_spider(project_id, spider_id)
    #     # -----所指定的 spider不存在-----
    #     if not spider:
    #         return Execution.NO_SPIDER
    #     # -----若 spider已完成-----
    #     if spider['status'] == self.tag_finished():
    #         logger.warning(f"[{project_id}/{spider_id}] 已完成，無須中斷爬蟲")
    #         return Execution.FAIL
    #     # -----若 spider待命中-----
    #     if spider['status'] == self.tag_stand_by():
    #         logger.warning(f"[{project_id}/{spider_id}] 待命中，無須中斷爬蟲")
    #         return Execution.FAIL
    #     # -----若 spider停止中-----
    #     if spider['status'] == self.tag_stop():
    #         logger.warning(f"[{project_id}/{spider_id}] 停止中，無須中斷爬蟲")
    #         return Execution.FAIL
    #     # -----若 spider進行中-----
    #     if spider['status'] == self.tag_running():
    #         logger.info(f"[{project_id}/{spider_id}] 進行中，即將中斷爬蟲")
    #
    #     spider['jobs'].get(spider['job']).cancel()
    #     spider['status'] = self.tag_stop()
    #     spider['items'] = None
    #     spider['logs'] = None
    #
    #     return Execution.SUCCESS
    #
    # def resume_spider(self, project_id, spider_id):
    #
    #     self.make_sure_initialized()
    #     spider = self.check_and_get_spider(project_id, spider_id)
    #     # -----所指定的 spider不存在-----
    #     if not spider:
    #         return Execution.NO_SPIDER
    #     # -----若 spider待命中-----
    #     if spider['status'] == self.tag_stand_by():
    #         logger.warning(f"[{project_id}/{spider_id}] 待命中，無法恢復待命狀態")
    #         return Execution.FAIL
    #     # -----若 spider停止中-----
    #     if spider['status'] == self.tag_stop():
    #         logger.warning(f"[{project_id}/{spider_id}] 停止中，無法恢復待命狀態")
    #         return Execution.FAIL
    #     # -----若 spider進行中-----
    #     if spider['status'] == self.tag_running():
    #         logger.warning(f"[{project_id}/{spider_id}] 進行中，無法恢復待命狀態")
    #         return Execution.FAIL
    #     # -----若 spider已完成-----
    #     if spider['status'] == self.tag_finished():
    #         logger.info(f"[{project_id}/{spider_id}] 已完成，改為待命中狀態，並銷毀暫存資料實體")
    #         if spider["name"] == spider_id:
    #             spider['status'] = self.tag_stand_by()
    #             spider['items'] = None
    #             spider['logs'] = None
    #         return Execution.SUCCESS
    #
    # def retrieve_item(self, project_id, spider_id):
    #
    #     self.make_sure_initialized()
    #     spider = self.check_and_get_spider(project_id, spider_id)
    #     # -----所指定的 spider不存在-----
    #     if not spider:
    #         return None
    #     # -----若 spider待命中-----
    #     if spider['status'] == self.tag_stand_by():
    #         logger.warning(f"[{project_id}/{spider_id}] 待命中，不存在可供搬運資料")
    #         return None
    #     # -----若 spider停止中-----
    #     if spider['status'] == self.tag_stop():
    #         logger.warning(f"[{project_id}/{spider_id}] 停止中，不存在可供搬運資料")
    #         return None
    #     # -----若 spider進行中-----
    #     if spider['status'] == self.tag_running():
    #         logger.warning(f"[{project_id}/{spider_id}] 進行中，尚不提供可供搬運資料")
    #         return None
    #     # -----若 spider已完成-----
    #     if spider['status'] == self.tag_finished():
    #         logger.info(f"[{project_id}/{spider_id}] 已完成，將資料實體轉移至應用程式使用")
    #         job = spider.get('job', None)
    #         jobs = spider.get('jobs', None)
    #         if not job or not jobs:
    #             raise Exception("爬蟲已完成，無對應工作，程式邏輯異常")
    #         items = jobs.get(job).items
    #         if not items:
    #             raise Exception("爬蟲已完成，無可供操作實體，程式邏輯異常")
    #         return items

    # 以下為 spider mode 的功能
    # def add_spider(self, add_spider):
    #     self.make_sure_initialized()
    #     self.make_sure_spider_mode()
    #
    #     # 若傳入非 spider base物件則報錯
    #     if not isinstance(add_spider, SpiderBase):
    #         raise Exception("傳入非 spider_base 之子物件")
    #     if not self.check_and_get_spider(add_spider.project_id, add_spider.spider_id):
    #         raise Exception("傳入 spider 非紀錄之spider")
    #
    #     if not add_spider.tag:
    #         add_spider.add_tag(self.spiders_tag)
    #     # 若 add 相同 spider id 和 project id 的 spider，刪除並從後方插入
    #     index = 0
    #     for spider in self.spiders:
    #         if spider.project_id == add_spider.project_id and spider.spider_id == add_spider.spider_id:
    #             del self.spiders[index]
    #         else:
    #             index += 1
    #     self.spiders.append(add_spider)
    #     logger.info(
    #         f"[Spider Mode]: 已添加 {add_spider.project_id}/{add_spider.spider_id}, 序列:{len(self.spiders)}, 標籤: {add_spider.tag}")
    #
    # def wake_spider(self):
    #     self.make_sure_initialized()
    #     self.make_sure_spider_mode()
    #
    #     spider = self.spiders[self.spiders_mode_index]
    #
    #     light = self.check_spider(spider.project_id, spider.spider_id)
    #
    #     next_round = 0 if self.spiders_mode_index + 1 >= len(self.spiders) else self.spiders_mode_index + 1
    #
    #     self.spiders_mode_index = next_round
    #
    #     spider.react(light, next_round, self)
    #
    # def start_spider(self):
    #     self.make_sure_initialized()
    #     self.make_sure_spider_mode()
    #     # 並未加入爬蟲
    #     if len(self.spiders) == 0:
    #         logger.warning("Didn't add any spider into spiders. Cancel starting!")
    #         return
    #     logger.debug("try to start spider mode scraping")
    #     # 將當前爬蟲模式的次序歸零
    #     # new
    #     self.spiders_mode_index = 0
    #
    #     leader_spider = self.spiders[self.spiders_mode_index]
    #
    #     if leader_spider.shall_start(self):
    #         logger.debug("trying success. Spider mode scraping started")
    #         self.wake_spider()
    #
    #     else:
    #         logger.debug("trying fail. Spider mode scraping cannot pass 'shall_start()' restriction.")
    #     # old (v0.0.1)
    #     # index = self.spiders_mode_index + 1
    #     # if index >= len(self.spiders):
    #     #     index = 0
    #     # spider = self.spiders[index]
    #     # if spider.shall_start(self):
    #     #     self.wake_spider()
    #
    # def set_spiders_mode_index(self, order):
    #
    #     self.make_sure_initialized()
    #     self.make_sure_spider_mode()
    #
    #     if 0 <= order <= len(self.spiders) - 1:
    #         self.spiders_mode_index = order
    #         return True
    #     raise Exception("Setting index is out of spiders range")
    #
    # def check_spiders_status(self):
    #
    #     self.make_sure_initialized()
    #     self.make_sure_spider_mode()
    #
    #     spiders_status = {}
    #     index = 0
    #     for spider_item in self.spiders:
    #         spider = self.check_and_get_spider(spider_item.project_id, spider_item.spider_id)
    #         if not spider:
    #             raise Exception("自動爬蟲模式出現未被紀錄之spider，程式邏輯異常")
    #         spiders_status[f"{spider_item.project_id}/{spider_item.spider_id}"] = {
    #             "name": spider.get('name', None),
    #             "index": index,
    #             "status": spider.get('status', None)
    #         }
    #         index += 1
    #     return spiders_status
    #
    # def check_spiders(self):
    #
    #     self.make_sure_initialized()
    #     self.make_sure_spider_mode()
    #
    #     for spider in self.spiders:
    #         if not self.check_spider(spider.project_id, spider.spider_id):
    #             raise Exception("存在未被記錄之spider，程式邏輯異常")
    #
    # def make_sure_spider_mode(self):
    #     if not self.spiders_mode:
    #         raise Exception("Not turn on spider mode.")
    #
    # @staticmethod
    # def get_job_status(jobs, job):
    #     return jobs.get(job).metadata.get('state')

    # @classmethod
    # def tag_stand_by(cls):
    #     return "stand by"
    #
    # @classmethod
    # def tag_running(cls):
    #     return "running"
    #
    # @classmethod
    # def tag_finished(cls):
    #     return "finished"
    #
    # @classmethod
    # def tag_stop(cls):
    #     return "stop"



import setuptools

INSTALL_REQUIRES = ["scrapinghub==2.3.1", "msgpack-python==0.5.6", "msgpack==1.0.2"]

setuptools.setup(
    name="flask-scrapinghub",
    version="0.0.8",
    author="chienaeae",
    authror_email="chienaeae@gmail.com",
    description="python-scrapinghub with flask",
    long_description=
    '''
    # 開發中版本 ALPHA version
    
    python-scrapinghub with flask
    用於flask框架所開發的應用程式插件，可以快速對架設於scrapinghub的unit 進行指派，並監控完成度，直至完成後可定義下載資料至應用程式進行之客製化行為。
    ''',
    long_description_content_type="text/markdown",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3.9",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    install_requires=INSTALL_REQUIRES,
    python_requires='>=3.6',
)
